# --- FUNCTIONS --- #

# Function that finds markers with extrme segregation distortion relative to its neighbours
# d=data, w=width (number of markers), z=value that determines what is an "outlier" 
# can unhash plot to look at the markers identified

SD_outlier_finder <- function(d,w,z) {
  
  #only consider LGs with as many or more markers than the width defined
  names <- rownames(summaryMap(d)[summaryMap(d)$n.mar>w-1,])
  outliers <- vector()
  for (i in names[1:(length(names)-1)]) {
    CHR<-i
    temp1 <- statMark(d, stat.type="marker", chr=CHR)$marker$AB
    temp2 <- rollapply(temp1, w, function(x) mean(x), align="center")
    temp3 <- rollapply(temp1, w, function(x) sd(x), align="center")
    temp2 <- c(rep(head(temp2,1), (w-1)/2), temp2, rep(tail(temp2,1), (w-1)/2))
    temp3 <- c(rep(head(temp3,1), (w-1)/2), temp3, rep(tail(temp3,1), (w-1)/2))
    temp4 <- abs(temp1-temp2)>z*temp3
    #plot(temp1 ~ c(1:length(temp1)), col=as.numeric(temp4)+1, main=CHR)
    outliers <- c(outliers, markernames(d, chr=CHR)[temp4])
  }
  return(outliers)
}

# Function that iteratively removes dxos unitl there are no more
# d=data, x=number of crossovers

dxo_remover <- function(d,ndxo,byCHR,p) {
  success <- FALSE
  while (!success) {
    todrop <- markernames(d)[statMark(d, stat.type="marker")$marker$dxo>ndxo]
    d <- drop.markers(d, todrop)
    d <- mstmap(d, id="NA.", bychr=byCHR, dist.fun="kosambi", trace=TRUE, p.value=p, noMap.size = 2, noMap.dist = 5)
    success <- length(todrop)<1
  }
  return(d)
}

# Function that iteratively removes extreme sd unitl there is no more
# d=data, x=number of crossovers

sd_outlier_remover <- function(d,w,z,byCHR,p) {
  success <- FALSE
  while (!success) {
    todrop <- SD_outlier_finder(d,w,z)
    d <- drop.markers(d, todrop)
    d <- mstmap(d, id="NA.", bychr=byCHR, dist.fun="kosambi", trace=TRUE, p.value=p, noMap.size = 2, noMap.dist = 5)
    success <- length(todrop)<1
  }
  return(d)
}

# Function that finds Ann cM based on marker position which is found in the marker names

add_ANN_cM <- function(map) {
  
  library(stringr)
  
  # Add Ann chromosome name and position by splitting the marker names 
  map$AnnCHR <- str_split_fixed(map$markers, "_", 2)[,1]
  map$Annbp <- as.numeric(str_split_fixed(map$markers, "_", 2)[,2])
  
  # Remove extra ANN CHRs
  goodANNchrs <- c("HanXRQChr01", "HanXRQChr02", "HanXRQChr03", "HanXRQChr04", "HanXRQChr05", "HanXRQChr06", "HanXRQChr07", "HanXRQChr08", "HanXRQChr09", "HanXRQChr10", "HanXRQChr11", "HanXRQChr12", "HanXRQChr13", "HanXRQChr14", "HanXRQChr15", "HanXRQChr16", "HanXRQChr17")
  map <- map[which(map$AnnCHR %in% goodANNchrs),]
  
  # Load Ann cM to bp data
  ann.data <- read.table("04_data/XRQ_bp_cM.txt")
  names(ann.data) <- c("CHR", "bp1", "cM1", "bp2", "cM2")
  ann.data$bp1 <- as.numeric(ann.data$bp1)
  ann.data$bp2 <- as.numeric(ann.data$bp2)
  
  # Find upper and lower cM and estimate Ann cM
  for(i in 1:nrow(map)){
    map$AnncM1[i] <- ann.data[(ann.data$CHR==map$AnnCHR[i] & ann.data$bp1<=map$Annbp[i] & ann.data$bp2>map$Annbp[i]),3]
    map$AnncM2[i] <- ann.data[(ann.data$CHR==map$AnnCHR[i] & ann.data$bp1<=map$Annbp[i] & ann.data$bp2>map$Annbp[i]),5]
  }
  map$dist <- as.numeric(str_sub(map$Annbp, -5))
  map$AnncM <- map$AnncM1+((map$AnncM2-map$AnncM1)*(map$dist/100000))
  
  # Make new table with the appropriate columns
  map2 <- map[,c(1, 4, 9, 1:3)]
  names(map2) <- c("name1", "chr1", "pos1", "name2", "chr2", "pos2")
  
  return(map2)
  
}

add_ANN_cM2 <- function(map) {
  
  library(stringr)
  
# Remove extra ANN CHRs
  goodANNchrs <- c("HanXRQChr01", "HanXRQChr02", "HanXRQChr03", "HanXRQChr04", "HanXRQChr05", "HanXRQChr06", "HanXRQChr07", "HanXRQChr08", "HanXRQChr09", "HanXRQChr10", "HanXRQChr11", "HanXRQChr12", "HanXRQChr13", "HanXRQChr14", "HanXRQChr15", "HanXRQChr16", "HanXRQChr17")
  map <- map[which(map$chr %in% goodANNchrs),]
  map$chr <- factor(map$chr)
  
  # Load Ann cM to bp data
  ann.data <- read.table("04_data/XRQ_bp_cM.txt")
  names(ann.data) <- c("CHR", "bp1", "cM1", "bp2", "cM2")
  ann.data$bp1 <- as.numeric(ann.data$bp1)
  ann.data$bp2 <- as.numeric(ann.data$bp2)
  
  # Find upper and lower cM and estimate Ann cM
  for(i in 1:nrow(map)){
    map$AnncM1[i] <- ann.data[(ann.data$CHR==map$chr[i] & ann.data$bp1<=map$pos[i] & ann.data$bp2>map$pos[i]),3]
    map$AnncM2[i] <- ann.data[(ann.data$CHR==map$chr[i] & ann.data$bp1<=map$pos[i] & ann.data$bp2>map$pos[i]),5]
  }
  map$dist <- as.numeric(str_sub(map$pos, -5))
  map$AnncM <- map$AnncM1+((map$AnncM2-map$AnncM1)*(map$dist/100000))
  
  # Make new table with the appropriate columns
  map2 <- map[,c(1, 2, 7)]
  names(map2) <- c("marker", "chr", "pos")
  
  return(map2)
  
}

# function to add translocation/inversion color scheme to block data_frames

add_rearrangement_type_col <- function(synteny_block_df){
  
  synteny_block_df <- synteny_block_df %>% 
    mutate(chr1 = as.numeric(gsub("A", "", map1_chr)), 
           chr2 = floor(as.numeric(gsub("[A-z]", "", map2_chr)))) %>%
    mutate(col = case_when(
      chr1 == chr2 & orientation == "positive" ~ "grey",
      chr1 == chr2 & orientation == "unclassified" ~ "grey40",
      chr1 == chr2 & orientation == "negative" ~ "black",
      chr1 != chr2 & orientation == "negative" ~ "blue4",
      chr1 != chr2 & orientation == "unclassified" ~ "blue",
      TRUE ~ "cornflowerblue"))
  
  return(synteny_block_df)
}


# function to find common breaks between the 4 species, and add breaks unique to each species

find_min_breaks <- function(neg_int, pet_int, niv_int, arg_int){
  
  # find common breaks (full overlaps from triples, triples from pairs)
  full <- interval_intersection(neg_int, pet_int, niv_int, arg_int)
  triplet_1 <- interval_intersection(neg_int, pet_int, niv_int)
  triplet_1 <- triplet_1[!lengths(interval_overlap(triplet_1, full)) > 0]
  triplet_2 <- interval_intersection(neg_int, pet_int, arg_int)
  triplet_2 <- triplet_2[!lengths(interval_overlap(triplet_2, full)) > 0]
  triplet_3 <- interval_intersection(neg_int, niv_int, arg_int)
  triplet_3 <- triplet_3[!lengths(interval_overlap(triplet_3, full)) > 0]
  triplet_4 <- interval_intersection(pet_int, niv_int, arg_int)
  triplet_4 <- triplet_4[!lengths(interval_overlap(triplet_4, full)) > 0]
  pair_1 <- interval_intersection(neg_int, pet_int)
  pair_1 <- pair_1[!(lengths(interval_overlap(pair_1, triplet_1)) > 0)]
  pair_1 <- pair_1[!(lengths(interval_overlap(pair_1, triplet_2)) > 0)]
  pair_1 <- pair_1[!(lengths(interval_overlap(pair_1, triplet_3)) > 0)]
  pair_1 <- pair_1[!(lengths(interval_overlap(pair_1, triplet_4)) > 0)]
  pair_1 <- pair_1[!(lengths(interval_overlap(pair_1, full)) > 0)]
  pair_2 <- interval_intersection(neg_int, niv_int)
  pair_2 <- pair_2[!(lengths(interval_overlap(pair_2, triplet_1)) > 0)]
  pair_2 <- pair_2[!(lengths(interval_overlap(pair_2, triplet_2)) > 0)]
  pair_2 <- pair_2[!(lengths(interval_overlap(pair_2, triplet_3)) > 0)]
  pair_2 <- pair_2[!(lengths(interval_overlap(pair_2, triplet_4)) > 0)]
  pair_2 <- pair_2[!(lengths(interval_overlap(pair_2, full)) > 0)]
  pair_3 <- interval_intersection(neg_int, arg_int)
  pair_3 <- pair_3[!(lengths(interval_overlap(pair_3, triplet_1)) > 0)]
  pair_3 <- pair_3[!(lengths(interval_overlap(pair_3, triplet_2)) > 0)]
  pair_3 <- pair_3[!(lengths(interval_overlap(pair_3, triplet_3)) > 0)]
  pair_3 <- pair_3[!(lengths(interval_overlap(pair_3, triplet_4)) > 0)]
  pair_3 <- pair_3[!(lengths(interval_overlap(pair_3, full)) > 0)]
  pair_4 <- interval_intersection(pet_int, niv_int)
  pair_4 <- pair_4[!(lengths(interval_overlap(pair_4, triplet_1)) > 0)]
  pair_4 <- pair_4[!(lengths(interval_overlap(pair_4, triplet_2)) > 0)]
  pair_4 <- pair_4[!(lengths(interval_overlap(pair_4, triplet_3)) > 0)]
  pair_4 <- pair_4[!(lengths(interval_overlap(pair_4, triplet_4)) > 0)]
  pair_4 <- pair_4[!(lengths(interval_overlap(pair_4, full)) > 0)]
  pair_5 <- interval_intersection(pet_int, arg_int)
  pair_5 <- pair_5[!(lengths(interval_overlap(pair_5, triplet_1)) > 0)]
  pair_5 <- pair_5[!(lengths(interval_overlap(pair_5, triplet_2)) > 0)]
  pair_5 <- pair_5[!(lengths(interval_overlap(pair_5, triplet_3)) > 0)]
  pair_5 <- pair_5[!(lengths(interval_overlap(pair_5, triplet_4)) > 0)]
  pair_5 <- pair_5[!(lengths(interval_overlap(pair_5, full)) > 0)]
  pair_6 <- interval_intersection(niv_int, arg_int)
  pair_6 <- pair_6[!(lengths(interval_overlap(pair_6, triplet_1)) > 0)]
  pair_6 <- pair_6[!(lengths(interval_overlap(pair_6, triplet_2)) > 0)]
  pair_6 <- pair_6[!(lengths(interval_overlap(pair_6, triplet_3)) > 0)]
  pair_6 <- pair_6[!(lengths(interval_overlap(pair_6, triplet_4)) > 0)]
  pair_6 <- pair_6[!(lengths(interval_overlap(pair_6, full)) > 0)]
  
  common <- interval_union(full, triplet_1, triplet_2, triplet_3, triplet_4, pair_1, pair_2, pair_3, pair_4, pair_5, pair_6)

  neg_unique_int <- neg_int[lengths(interval_overlap(neg_int, common)) == 0]
  pet_unique_int <- pet_int[lengths(interval_overlap(pet_int, common)) == 0]
  niv_unique_int <- niv_int[lengths(interval_overlap(niv_int, common)) == 0]
  arg_unique_int <- arg_int[lengths(interval_overlap(arg_int, common)) == 0]
  
  all <- interval_union(common, neg_unique_int, pet_unique_int, niv_unique_int, arg_unique_int)
  
  return(all)
  
}

# funtion to find orientation estimates from final blocks when there is master list out output data dervided from multiple k values

estimate_orientation <- function(final_blocks, output) {
  
  k_s <- final_blocks[final_blocks$orientation == "unclassified",]$k
  block_n_s <- final_blocks[final_blocks$orientation == "unclassified",]$block
  
  orientation_list <- vector()
  
  for(i in 1:length(k_s)){
    
    k <- as.numeric(k_s[i])
    block_n <- as.numeric(block_n_s[i])
    xs <- output[[k]][[1]][output[[k]][[1]]$block == block_n,]$map1_pos
    ys <- output[[k]][[1]][output[[k]][[1]]$block == block_n,]$map2_pos
    temp <- summary(lm(ys ~ xs))[[4]][2]>0
    orientation_list <- c(orientation_list, temp)
    
  }
  
  return(ifelse(orientation_list == T, "positive", "negative"))
  
}

# funtion to find orientation estimates for final blocks for regular syntR output

estimate_orientation_basic <- function(syntr_all_dfs){
  
  best_guess_orientations <- vector()
  
  for(i in syntr_all_dfs[[2]]$block){
    
    xs <- syntr_all_dfs[[1]] %>% filter(final_block == i) %>% select(map1_pos) %>% unlist
    ys <- syntr_all_dfs[[1]] %>% filter(final_block == i) %>% select(map2_pos) %>% unlist
    temp <- summary(lm(ys ~ xs))[[4]][2]>0
    best_guess_orientations <- c(best_guess_orientations, temp)
  
  }
  
  return(ifelse(best_guess_orientations == T, "positive", "negative"))
  
}
  

# function to plot positive, negative and unclassified segments from a block list

plot_block_orientations <- function(blocks){
  
  # plot segments
  segs1 <- blocks %>% filter(orientation == "positive") 
  for(j in 1:nrow(segs1)) {
    segments(x0 = segs1$map1_start[j], y0 = segs1$map2_start[j], x1 = segs1$map1_end[j], y1 = segs1$map2_end[j], col = "blue", lwd = 3)
  }
  segs2 <- blocks %>% filter(orientation == "negative") 
  for(j in 1:nrow(segs2)) {
    segments(x0 = segs2$map1_start[j], y0 = segs2$map2_end[j], x1 = segs2$map1_end[j], y1 = segs2$map2_start[j], col = "red", lwd = 3)
  }
  segs3 <- blocks %>% filter(orientation == "unclassified") 
  for(j in 1:nrow(segs3)) {
    segments(x0 = segs3$map1_start[j], y0 = segs3$map2_start[j], x1 = segs3$map1_end[j], y1 = segs3$map2_end[j], col = "purple", lwd = 3)
    segments(x0 = segs3$map1_start[j], y0 = segs3$map2_end[j], x1 = segs3$map1_end[j], y1 = segs3$map2_start[j], col = "purple", lwd = 3)
  }
  
}